@@ -1,40 +1,4 @@
-/* Notes *//*{{{C}}}*//*{{{*/
-/*
-
-This modules implements generic functions which work on a gap buffer of
-character and meta-characters.  A gap buffer is a common way for to
-store text in text editors and it works like this:
-
-| text before gap |   ... gap ... | text after gap |
-
-The start/end of the gap reflects the cursor position in the text.
-Inserting is done by decreasing the gap, deleting by increasing it.  The
-stored text consists of characters and fold marks.  The special
-character '\0' quotes a following fold mark or nul character, using the
-constants NUL, FOLDSTART and FOLDEND.  Special care must be taken not to
-have more than one fold mark per line and to keep the fold structure 
-when using the functions from this module.
-
-My personal experience from hacking on the folding editor Origami is, 
-that a fancy data structure (double-linked n-tree) is very efficient, as
-even with a deeply folded 3 MB file there is no slowdown.  This can not 
-be expected from using a gap buffer.  The reason why I use it never the 
-less, is that a fancy data structure makes it nearly impossible to have 
-correct and elegant primitives which work on the raw text.  In fact, 
-a significant slow down can be felt from lots of node-traversing which 
-happens when deleting bigger amounts of text.  This is the reason, why 
-most versions of Origami enforce using folds for everything.  While this
-seemed elegant in the beginning, after a longer time I have found that 
-folds are no universal solution to everything.
-
-As a result, this implementation is slower for basic operations (opening
-and closing folds, screen display), faster for complex operations (block
-operations, loading, saving) and tries to offer the best of both worlds
-(traditional text editor and folding editor).
-
-*/
-/*}}}*/
-/* #includes *//*{{{*/
+/* #includes *//*{{{C}}}*//*{{{*/
 #ifndef NO_POSIX_SOURCE
 #undef  _POSIX_SOURCE
 #define _POSIX_SOURCE   1
@@ -66,7 +30,7 @@ operations, loading, saving) and tries to offer the best of both worlds
 /* #defines *//*{{{*/
 #define DEF_BUFLEN 1024
 
-#if 0
+#if 1
 #define ASSERT_CONSISTENCY(b) \
 assert(b!=(struct Buffer*)0); \
 assert(b->buffer!=(char*)0); \
@@ -78,37 +42,65 @@ assert(b->gapend<=b->buffer+b->size-1)
 
 #define BF_FORWARDOK(b) (b->gapend<b->buffer+b->size-1)
 
+/*
+In UTF-8 mode the first char encodes the length:
+110xxxxx = 2 bytes
+1110xxxx = 3 bytes
+11110xxx = 4 bytes
+*/
+
 #define BF_FORWARD(b) \
-switch (++b->curch,*(b->gapstart++)=*(++b->gapend)) \
+do \
 { \
-  case '\0': \
+  char bf_prev; \
+  ++b->curch; \
+  bf_prev=*(b->gapstart++)=*(++b->gapend); \
+  if (bf_prev=='\0' && *(b->gapend+1)==QUOTE) \
   { \
     *(b->gapstart++)=*(++b->gapend); \
-    break; \
+    *(b->gapstart++)=*(++b->gapend); \
   } \
-  case '\n': \
+  else if (bf_prev=='\0') *(b->gapstart++)=*(++b->gapend); \
+  else if (bf_prev=='\n') ++b->cury; \
+  else if (bf_utf8) \
   { \
-    ++b->cury; \
-    break; \
+    if ((bf_prev & 0xf0)==0xf0) *(b->gapstart++)=*(++b->gapend); \
+    if ((bf_prev & 0xe0)==0xe0) *(b->gapstart++)=*(++b->gapend); \
+    if ((bf_prev & 0xc0)==0xc0) *(b->gapstart++)=*(++b->gapend); \
   } \
-}
+} while (0)
 
 #define BF_BACKWARDOK(b) (b->gapstart>b->buffer)
 
 #define BF_BACKWARD(b) \
 do \
 { \
-  *(b->gapend--)=*(--b->gapstart); \
-  if (b->gapstart-1>=b->buffer && *(b->gapstart-1)=='\0') \
+  char bf_cur; \
+  bf_cur = *(b->gapend--)=*(--b->gapstart); \
+  if (b->gapstart>b->buffer && *(b->gapstart-1)=='\0') \
   { \
     *(b->gapend--)=*(--b->gapstart); \
   } \
-  if (*(b->gapend+1)=='\n') --b->cury; \
+  else if (b->gapstart-1>b->buffer && *(b->gapstart-2)=='\0' && *(b->gapstart-1)==QUOTE) \
+  { \
+    *(b->gapend--)=*(--b->gapstart); \
+    *(b->gapend--)=*(--b->gapstart); \
+  } \
+  else if (bf_cur=='\n') --b->cury; \
+  else if ((bf_utf8) && (bf_cur & 0x80)) \
+  { \
+    while ((bf_cur & 0xc0) != 0xc0) bf_cur = *(b->gapend--)=*(--b->gapstart); \
+  } \
   --b->curch; \
-} while(0);
+} while (0);
 /*}}}*/
 
 /* variables *//*{{{*/
+int bf_utf8;
+
+const struct MetaChar bf_foldstart = { { '\0', FOLDSTART } };
+const struct MetaChar bf_foldend = { { '\0', FOLDEND } };
+
 struct Language language[LANGUAGES]=
 {
   { "C", "/*", "*/" },
@@ -128,7 +120,7 @@ struct Language language[LANGUAGES]=
 };
 /*}}}*/
 
-/* bf_alloc      -- allocate new buffer *//*{{{*/
+/* bf_alloc        -- allocate new buffer *//*{{{*/
 struct Buffer *bf_alloc(void)
 {
   struct Buffer *b;
@@ -154,7 +146,7 @@ struct Buffer *bf_alloc(void)
   return b;
 }
 /*}}}*/
-/* bf_free       -- free buffer *//*{{{*/
+/* bf_free         -- free buffer *//*{{{*/
 void bf_free(struct Buffer *b)
 {
   ASSERT_CONSISTENCY(b);
@@ -165,10 +157,11 @@ void bf_free(struct Buffer *b)
   free(b);
 }
 /*}}}*/
-/* bf_empty      -- empty buffer */ /*{{{*/
+/* bf_empty        -- empty buffer */ /*{{{*/
 int bf_empty(struct Buffer *b)
 {
   int i;
+
   if (b->mode&READONLY) return 0;
   b->gapstart=b->buffer;
   b->gapend=b->buffer+b->size-1;
@@ -181,15 +174,20 @@ int bf_empty(struct Buffer *b)
   return 1;
 }
 /*}}}*/
-/* bf_forward    -- move one character forward *//*{{{*/
+/* bf_forward      -- move one character forward *//*{{{*/
 int bf_forward(struct Buffer *b)
 {
   ASSERT_CONSISTENCY(b);
-  if (BF_FORWARDOK(b)) { BF_FORWARD(b); return 1; }
+  if (BF_FORWARDOK(b))
+  {
+    BF_FORWARD(b);
+    ASSERT_CONSISTENCY(b);
+    return 1;
+  }
   else return 0;
 }
 /*}}}*/
-/* bf_backward   -- move one character backward *//*{{{*/
+/* bf_backward     -- move one character backward *//*{{{*/
 int bf_backward(struct Buffer *b)
 {
   ASSERT_CONSISTENCY(b);
@@ -197,15 +195,49 @@ int bf_backward(struct Buffer *b)
   else return 0;
 }
 /*}}}*/
-/* bf_insert     -- insert one character *//*{{{*/
-int bf_insert(struct Buffer *b, char c, char mark)
+/* bf_lcharlen     -- return length of ending MetaChar *//*{{{*/
+size_t bf_lcharlen(const char *s, const char *start)
 {
-  int i;
+  const char *end;
+
+  end=s;
+  if (s>start && *(s-1)=='\0') --s;
+  if (s-1>start && *(s-2)=='\0' && *(s-1)==QUOTE) s-=2;
+  else if (bf_utf8 && (*s & 0x80))
+  {
+    for (--s; (*s & 0xc0) != 0xc0; --s);
+  }
+  return end-s+1;
+}
+/*}}}*/
+/* bf_rcharlen     -- return length of starting MetaChar *//*{{{*/
+size_t bf_rcharlen(const char *s)
+{
+  if (*s=='\0')
+  {
+    return (*(s+1)==QUOTE ? 3 : 2);
+  }
+  else if (bf_utf8)
+  {
+    if (!(*s & 0x80)) return 1;
+    if ((*s & 0xe0) == 0xc0) return 2;
+    else if ((*s & 0xf0) == 0xe0) return 3;
+    else if ((*s & 0xf8) == 0xf0) return 4;
+    else assert(0);
+  }
+  else return 1;
+}
+/*}}}*/
+/* bf_insert       -- insert one character *//*{{{*/
+int bf_insert(struct Buffer *b, const struct MetaChar *ch)
+{
+  int i,len;
+  const char *cp;
 
   ASSERT_CONSISTENCY(b);
   if (b->mode&READONLY) return 0;
-  if (b->gapstart>=b->gapend-2)
-  /* increase gap *//*{{{*/
+
+  if (b->gapstart>=b->gapend-4) /* increase gap *//*{{{*/
   {
     char *new_buffer;
     size_t new_size,pregap_length,postgap_length;
@@ -228,30 +260,21 @@ int bf_insert(struct Buffer *b, char c, char mark)
     b->size=new_size;
     b->gapstart=b->buffer+pregap_length;
     b->gapend=b->buffer+b->size-1-postgap_length;
+    ASSERT_CONSISTENCY(b);
   }
   /*}}}*/
+
   for (i=0; i<b->msz; ++i) if (b->gapstart-b->buffer<=b->mark[i]) ++b->mark[i];
-  *(b->gapstart++)=c;
-  switch (c)
-  {
-    case '\0':
-    {
-      assert(mark!='\0');
-      *(b->gapstart++)=mark;
-      break;
-    }
-    case '\n':
-    {
-      ++b->cury;
-      break;
-    }
-  }
+
+  cp=ch->c;
+  if (*cp=='\n') ++b->cury;
+  for (i=0,len=bf_rcharlen(cp); i<len; ++i) *(b->gapstart++)=*cp++;
   b->mode|=CHANGED;
   ++b->curch;
   return 1;
 }
 /*}}}*/
-/* bf_delete     -- delete one character *//*{{{*/
+/* bf_delete       -- delete one character *//*{{{*/
 int bf_delete(struct Buffer *b)
 {
   int i;
@@ -261,13 +284,12 @@ int bf_delete(struct Buffer *b)
   if (b->gapstart==b->buffer) return 0;
   else
   {
+    struct MetaChar ch;
+
     for (i=0; i<b->msz; ++i) if (b->gapstart-b->buffer<=b->mark[i]) --b->mark[i];
-    --b->gapstart;
-    if (b->gapstart-1>=b->buffer && *(b->gapstart-1)=='\0')
-    {
-      --b->gapstart;
-    }
-    else if (*(b->gapstart)=='\n')
+
+    b->gapstart-=bf_lchar(b,&ch);
+    if (ASCII(ch)=='\n')
     {
       --b->cury;
     }
@@ -277,36 +299,120 @@ int bf_delete(struct Buffer *b)
   }
 }
 /*}}}*/
-/* bf_lchar      -- return character before beginning of gap *//*{{{*/
-int bf_lchar(struct Buffer *b, char *c, char *mark)
+/* bf_insertstr    -- insert string of characters (external, no MetaChars) *//*{{{*/
+int bf_insertstr(struct Buffer *b, const char *str, const char *end)
 {
+  assert(str<=end);
+
+  while (str<end)
+  {
+    struct MetaChar ch;
+
+    if (*str=='\0')
+    {
+      ch.c[0] = '\0';
+      ch.c[1] = NUL;
+      ++str;
+    }
+    else if (bf_utf8)
+    {
+      size_t len;
+
+      if (!(*str & 0x80))
+      {
+        ch.c[0] = *str++;
+      }
+      else if ((*str & 0xe0) == 0xc0 && (str+1<end) && ((*(str+1)&0xc0) == 0x80))
+      {
+        ch.c[0] = *str++;
+        ch.c[1] = *str++;
+      }
+      else if ((*str & 0xf0) == 0xe0 && (str+2<end) && ((*(str+1)&0xc0) == 0x80) && ((*(str+2)&0xc0) == 0x80))
+      {
+        ch.c[0] = *str++;
+        ch.c[1] = *str++;
+        ch.c[2] = *str++;
+      }
+      else if ((*str & 0xf8) == 0xf0 && (str+3<end) && ((*(str+1)&0xc0) == 0x80) && ((*(str+2)&0xc0) == 0x80) && ((*(str+3)&0xc0) == 0x80))
+      {
+        ch.c[0] = *str++;
+        ch.c[1] = *str++;
+        ch.c[2] = *str++;
+        ch.c[3] = *str++;
+      }
+      else
+      {
+        /* Invalid UTF-8 character */
+        ch.c[0] = '\0';
+        ch.c[1] = QUOTE;
+        ch.c[2] = *str++;
+      }
+    }
+    else
+    {
+      ch.c[0] = *str++;
+    }
+
+    if (!bf_insert(b,&ch)) return 0;
+  }
+  return 1;
+}
+/*}}}*/
+/* bf_lchar        -- get character before beginning of gap, return length *//*{{{*/
+int bf_lchar(const struct Buffer *b, struct MetaChar *ch)
+{
+  const char *s;
+
   ASSERT_CONSISTENCY(b);
   if (b->gapstart==b->buffer) return 0;
-  if (b->gapstart-2>=b->buffer && *(b->gapstart-2)=='\0')
+
+  s=b->gapstart-1;
+  if (s-2>=b->buffer && *(s-2)=='\0' && *(s-1)==QUOTE) s-=2;
+  else if (s-1>=b->buffer && *(s-1)=='\0') --s;
+  else if (bf_utf8 && (*s & 0x80))
   {
-    *c='\0';
-    *mark=*(b->gapstart-1);
+    for (--s; (*s & 0xc0) != 0xc0; --s);
   }
-  else *c=*(b->gapstart-1);
-  return 1;
+  assert(b->gapstart-s<=4);
+  memcpy(ch->c, s, b->gapstart-s);
+  return b->gapstart-s;
 }
 /*}}}*/
-/* bf_rchar      -- return character after end of gap *//*{{{*/
-int bf_rchar(struct Buffer *b, char *c, char *mark)
+/* bf_rchar        -- get character after end of gap, return length *//*{{{*/
+int bf_rchar(const struct Buffer *b, struct MetaChar *ch)
 {
+  const char *s;
+
   ASSERT_CONSISTENCY(b);
+  assert(ch);
+
   if (b->gapend==b->buffer+b->size-1) return 0;
-  if ((*c=*(b->gapend+1))=='\0') *mark=*(b->gapend+2);
-  return 1;
+
+  s=b->gapend+1;
+  if (*s=='\0')
+  {
+    ++s;
+    if (*s==QUOTE) ++s;
+  }
+  else if (bf_utf8)
+  {
+    if ((*s & 0xe0) == 0xc0) ++s;
+    else if ((*s & 0xf0) == 0xe0) s+=2;
+    else if ((*s & 0xf8) == 0xf0) s+=3;
+  }
+
+  assert(s-b->gapend<=4);
+  memcpy(ch->c, b->gapend+1, s-b->gapend);
+  return s-b->gapend;
 }
 /*}}}*/
-/* bf_name       -- set buffer name *//*{{{*/
+/* bf_name         -- set buffer name *//*{{{*/
 int bf_name(struct Buffer *b, const char *name)
 {
   char *n;
 
   if (b->name) free(b->name);
-  if ((n=malloc(strlen(name)+1))==(char *)0) return 0;
+  if ((n=malloc(strlen(name)+1))==(char*)0) return 0;
   strcpy(n,name);
   b->name=n;
   b->mode|=CHANGED;
@@ -314,7 +420,7 @@ int bf_name(struct Buffer *b, const char *name)
   return 1;
 }
 /*}}}*/
-/* bf_lang       -- get/set buffer language *//*{{{*/
+/* bf_lang         -- get/set buffer language *//*{{{*/
 struct Language *bf_lang(struct Buffer *b, struct Language *l)
 {
   if (l)
@@ -326,7 +432,7 @@ struct Language *bf_lang(struct Buffer *b, struct Language *l)
   return b->lang;
 }
 /*}}}*/
-/* bf_load       -- load buffer from file *//*{{{*/
+/* bf_load         -- load buffer from file *//*{{{*/
 int bf_load(struct Buffer *b, const char *cmd, const char *file, int filefd, const char *beg, const char *end)
 {
   /* variables *//*{{{*/
@@ -396,9 +502,11 @@ int bf_load(struct Buffer *b, const char *cmd, const char *file, int filefd, con
   do
   {
     /* enlarge read buffer if needed *//*{{{*/
-    if (linerun==line+linesize)
+    if (linerun+1>=line+linesize)
     {
-      if ((line=realloc(line,linesize+256))==(char*)0)
+      char *larger_line;
+
+      if ((larger_line=realloc(line,linesize+256))==(char*)0)
       /* can not allocate more memory, close file and return failure *//*{{{*/
       {
         int oerrno;
@@ -413,7 +521,8 @@ int bf_load(struct Buffer *b, const char *cmd, const char *file, int filefd, con
       /*}}}*/
       else
       {
-        linerun=line+linesize;
+        linerun=larger_line+(linerun-line);
+        line=larger_line;
         linesize+=256;
       }    
     }
@@ -424,11 +533,15 @@ int bf_load(struct Buffer *b, const char *cmd, const char *file, int filefd, con
     --remaining;
     if ((eof || (*linerun++=*bufrun++)=='\n') && linerun>line) /* process line */
     {
+      size_t len;
+
+      *linerun='\0'; /* invalidates incomplete UTF-8 sequence */
+      len=linerun-line;
       if (search_for==LANGUAGE) /* search for magic mark *//*{{{*/
       {
-        const char *found_beg,*found_end;
+        char *found_beg,*found_end;
 
-        if ((found_beg=mymemmem(line,linerun-line,beg,begsz))!=(char*)0 && (found_end=mymemmem(found_beg+begsz,linerun-found_beg-begsz,end,endsz))!=(char*)0)
+        if ((found_beg=mymemmem(line,len,beg,begsz)) && (found_end=mymemmem(found_beg+begsz,linerun-found_beg-begsz,end,endsz)))
         {
           struct Language *langrun;
 
@@ -438,7 +551,7 @@ int bf_load(struct Buffer *b, const char *cmd, const char *file, int filefd, con
             if (strncmp(found_beg+begsz,langrun->name,strlen(langrun->name))==0)
             {
               /* variables */ /*{{{*/
-              const char *found;
+              char *found;
               /*}}}*/
 
               /* determine opening and closing fold marks */ /*{{{*/
@@ -447,21 +560,21 @@ int bf_load(struct Buffer *b, const char *cmd, const char *file, int filefd, con
               opening=malloc((openingsz=strlen(langrun->beg)+begsz+strlen(langrun->end))+1);
               strcpy(opening,langrun->beg);
               strcpy(opening+strlen(langrun->beg),beg);
-              strncpy(opening+strlen(langrun->beg)+begsz,langrun->end,strlen(langrun->end));
+              memcpy(opening+strlen(langrun->beg)+begsz,langrun->end,strlen(langrun->end));
               closing=malloc((closingsz=strlen(langrun->beg)+endsz+strlen(langrun->end))+1);
               strcpy(closing,langrun->beg);
               strcpy(closing+strlen(langrun->beg),end);
-              strncpy(closing+strlen(langrun->beg)+endsz,langrun->end,strlen(langrun->end));
+              memcpy(closing+strlen(langrun->beg)+endsz,langrun->end,strlen(langrun->end));
               /*}}}*/
               /* now check if this is also a opening fold */
               if ((found=mymemmem(found_end+endsz,linerun-found_end-endsz,opening,openingsz))!=(char*)0) /* good */ /*{{{*/
               {
-                const char *s;
-
-                for (s=line; s<found_beg-strlen(langrun->beg); ++s) bf_insert(b,*s,NUL);
-                for (s=found_end+endsz+strlen(langrun->end); s<found; ++s) bf_insert(b,*s,NUL);
-                bf_insert(b,'\0',FOLDSTART);
-                for (s=found+openingsz; s<linerun; ++s) bf_insert(b,*s,NUL);
+                *(found_beg-strlen(langrun->beg))='\0';
+                bf_insertstr(b,line,found_beg-strlen(langrun->beg));
+                *found='\0';
+                bf_insertstr(b,found_end+endsz+strlen(langrun->end), found);
+                bf_insert(b,&bf_foldstart);
+                bf_insertstr(b,found+openingsz,linerun);
                 search_for=FOLDS;
                 b->lang=langrun;
                 break;
@@ -469,42 +582,42 @@ int bf_load(struct Buffer *b, const char *cmd, const char *file, int filefd, con
               /*}}}*/
             }
           }
+
           if (langrun->name==(const char*)0) /* not found, so this was just a regular line */ /*{{{*/
           {
-            char *s;
-
-            for (s=line; s<linerun; ++s) bf_insert(b,*s,NUL);
+            
+            bf_insertstr(b,line,linerun);
           }
           /*}}}*/
           /*}}}*/
         }
         else
         {
-          const char *s;
-
-          for (s=line; s<linerun; ++s) bf_insert(b,*s,NUL);    
+          bf_insertstr(b,line,linerun);
         }
       }
       /*}}}*/
-      else /* search for folds *//*{{{*/
+      else /* search for folds or insert literally *//*{{{*/
       {
-        const char *found,*s;
+        char *found;
         
-        if ((found=mymemmem(line,linerun-line,opening,openingsz))!=(char*)0)
+        if ((found=mymemmem(line,len,opening,openingsz))!=(char*)0)
         {
-          for (s=line; s<found; ++s) bf_insert(b,*s,NUL);
-          bf_insert(b,'\0',FOLDSTART);
-          for (s=found+openingsz; s<linerun; ++s) bf_insert(b,*s,NUL);
+          *found='\0';
+          bf_insertstr(b,line,found);
+          bf_insert(b,&bf_foldstart);
+          bf_insertstr(b,found+openingsz,linerun);
         }
-        else if ((found=mymemmem(line,linerun-line,closing,closingsz))!=(char*)0)
+        else if ((found=mymemmem(line,len,closing,closingsz))!=(char*)0)
         {
-          for (s=line; s<found; ++s) bf_insert(b,*s,NUL);
-          bf_insert(b,'\0',FOLDEND);
-          for (s=found+closingsz; s<linerun; ++s) bf_insert(b,*s,NUL);
+          *found='\0';
+          bf_insertstr(b,line,found);
+          bf_insert(b,&bf_foldend);
+          bf_insertstr(b,found+openingsz,linerun);
         }
         else
         {
-          for (s=line; s<linerun; ++s) bf_insert(b,*s,NUL);
+          bf_insertstr(b,line,linerun);
         }
       }
       /*}}}*/
@@ -520,15 +633,15 @@ int bf_load(struct Buffer *b, const char *cmd, const char *file, int filefd, con
   return 1;
 }
 /*}}}*/
-/* bf_save       -- save buffer into file *//*{{{*/
+/* bf_save         -- save buffer into file *//*{{{*/
 int bf_save(struct Buffer *b, const char *file, const char *beg, const char *end, unsigned int *chars)
 {
   /* variables *//*{{{*/
   char *gapstart;
   char *bufp;
   char buf[1024];
-  char c,mark;
-  const char *writestr;
+  struct MetaChar ch;
+  size_t len;
   int fd;
   char *magic,*begfold,*endfold;
   int first=1;
@@ -542,7 +655,7 @@ int bf_save(struct Buffer *b, const char *file, const char *beg, const char *end
   strcpy(begfold,b->lang ? b->lang->beg : "");
   strcat(begfold,beg);
   if (b->lang) strcat(begfold,b->lang->end);
-  if ((magic=malloc(strlen(b->lang ? b->lang->beg : "")+strlen(beg)+strlen(b->lang ? b->lang->name : "")+strlen(end)+strlen(b->lang ? b->lang->end : "")+strlen(begfold)))==(char*)0)
+  if ((magic=malloc(strlen(b->lang ? b->lang->beg : "")+strlen(beg)+strlen(b->lang ? b->lang->name : "")+strlen(end)+strlen(b->lang ? b->lang->end : "")+strlen(begfold)+1))==(char*)0)
   {
     free(begfold);
     return 0;
@@ -553,7 +666,7 @@ int bf_save(struct Buffer *b, const char *file, const char *beg, const char *end
   strcat(magic,end);
   if (b->lang) strcat(magic,b->lang->end);
   strcat(magic,begfold);
-  if ((endfold=malloc(strlen(b->lang ? b->lang->beg : "")+strlen(end)+strlen(b->lang ? b->lang->end : "")))==(char*)0)
+  if ((endfold=malloc(strlen(b->lang ? b->lang->beg : "")+strlen(end)+strlen(b->lang ? b->lang->end : "")+1))==(char*)0)
   {
     free(begfold);
     free(magic);
@@ -564,22 +677,31 @@ int bf_save(struct Buffer *b, const char *file, const char *beg, const char *end
   if (b->lang) strcat(endfold,b->lang->end);
   if ((fd=open(file,O_WRONLY|O_CREAT|O_TRUNC,0666))==-1) return 0;
   gapstart=b->gapstart;
-  writestr=(const char*)0;
   bufp=buf;
   bf_begin(b);
-  while (bf_rchar(b,&c,&mark))
+  while ((len=bf_rchar(b,&ch)))
   {
-    if (c=='\0') switch (mark)
+    const char *writestr;
+    size_t writelen=0;
+
+    if (ASCII(ch)=='\0') switch (MARK(ch))
     {
-      case NUL: *bufp++='\0'; break;
-      case FOLDSTART: writestr=(first ? magic : begfold); first=0; break;
-      case FOLDEND: writestr=endfold; break;
+      case NUL: writestr=""; writelen=1; break;
+      case QUOTE: writestr=ch.c+2; writelen=1; break;
+      case FOLDSTART: writestr=(first ? magic : begfold); first=0; writelen=strlen(writestr); break;
+      case FOLDEND: writestr=endfold; writelen=strlen(writestr); break;
       default: assert(0);
     }
-    else *bufp++=c;
+    else
+    {
+      writestr=ch.c; writelen=len;
+    }
+
+    assert(writelen);
     do
     {
-      if (writestr && *writestr) *bufp++=*writestr++;
+      *bufp++=*writestr++;
+      --writelen;
       if (bufp==buf+sizeof(buf))
       {
         if (write(fd,buf,sizeof(buf))!=sizeof(buf))
@@ -597,10 +719,9 @@ int bf_save(struct Buffer *b, const char *file, const char *beg, const char *end
         else if (chars!=(unsigned int*)0) *chars+=sizeof(buf);
         bufp=buf;
       }
-    } while (writestr && *writestr);
+    } while (writelen);
     BF_FORWARD(b);
-  }
-  while (gapstart<b->gapstart) BF_BACKWARD(b);
+  } while (gapstart<b->gapstart) BF_BACKWARD(b);
   free(begfold);
   free(magic);
   free(endfold);
@@ -622,14 +743,15 @@ int bf_save(struct Buffer *b, const char *file, const char *beg, const char *end
   return 1;
 }
 /*}}}*/
-/* bf_write      -- write buffer without fold marks into file *//*{{{*/
+/* bf_write        -- write buffer without fold marks into file *//*{{{*/
 int bf_write(struct Buffer *b, const char *file, unsigned int *chars)
 {
   /* variables *//*{{{*/
   char *gapstart;
   char *bufp;
   char buf[1024];
-  char c,mark;
+  struct MetaChar ch;
+  size_t len;
   int fd;
   /*}}}*/
 
@@ -639,29 +761,44 @@ int bf_write(struct Buffer *b, const char *file, unsigned int *chars)
   gapstart=b->gapstart;
   bufp=buf;
   bf_begin(b);
-  while (bf_rchar(b,&c,&mark))
+  while ((len=bf_rchar(b,&ch)))
   {
-    if (c=='\0') switch (mark)
+    size_t writelen=0;
+    char const *writestr;
+
+    if (ASCII(ch)=='\0') switch (MARK(ch))
     {
-      case NUL: *bufp++='\0'; break;
+      case NUL: writestr=""; writelen+1; break;
+      case QUOTE: writestr=ch.c+2; writelen=1; break;
       case FOLDSTART: break;
       case FOLDEND: break;
       default: assert(0);
     }
-    else *bufp++=c;
-    if (bufp==buf+sizeof(buf))
+    else
     {
-      if (write(fd,buf,sizeof(buf))!=sizeof(buf))
+      writestr=ch.c;
+      writelen=len;
+    }
+
+    while (writelen)
+    {
+      *bufp++=*writestr++;
+      --writelen;
+
+      if (bufp==buf+sizeof(buf))
       {
-        int oerrno;
+        if (write(fd,buf,sizeof(buf))!=sizeof(buf))
+        {
+          int oerrno;
       
-        oerrno=errno;
-        close(fd);
-        errno=oerrno;
-        return 0;
+          oerrno=errno;
+          close(fd);
+          errno=oerrno;
+          return 0;
+        }
+        else if (chars!=(unsigned int*)0) *chars+=sizeof(buf);
+        bufp=buf;
       }
-      else if (chars!=(unsigned int*)0) *chars+=sizeof(buf);
-      bufp=buf;
     }
     BF_FORWARD(b);
   }
@@ -684,74 +821,77 @@ int bf_write(struct Buffer *b, const char *file, unsigned int *chars)
   return 1;
 }
 /*}}}*/
-/* bf_linestart  -- go to start of current line *//*{{{*/
+/* bf_linestart    -- go to start of current line *//*{{{*/
 int bf_linestart(struct Buffer *b, int realstart)
 {
-  char c,mark;
+  struct MetaChar ch;
   
+  ASSERT_CONSISTENCY(b);
   if (realstart)
   {
     if (b->gapstart==b->buffer) return 0;
-    while (bf_lchar(b,&c,&mark) && c!='\n' && BF_BACKWARDOK(b)) BF_BACKWARD(b);
+    while (bf_lchar(b,&ch) && ASCII(ch)!='\n' && BF_BACKWARDOK(b)) BF_BACKWARD(b);
   }
   else
   {
     int blanks;
 
     blanks=0;
-    if (bf_lchar(b,&c,&mark) && c!='\n')
+    if (bf_lchar(b,&ch) && ASCII(ch)!='\n')
     {
-      if (bf_rchar(b,&c,&mark) && c!=' ') blanks=1;
-      while (bf_lchar(b,&c,&mark) && c!='\n' && BF_BACKWARDOK(b))
+      if (bf_rchar(b,&ch) && ASCII(ch)!=' ') blanks=1;
+      while (bf_lchar(b,&ch) && ASCII(ch)!='\n' && BF_BACKWARDOK(b))
       {
         BF_BACKWARD(b);
-        if (c==' ' && blanks) ++blanks; else blanks=0;
+        if (ASCII(ch)==' ' && blanks) ++blanks; else blanks=0;
       }
     }
-    if (blanks==0) while (bf_rchar(b,&c,&mark) && c==' ') BF_FORWARD(b);
+    if (blanks==0) while (bf_rchar(b,&ch) && ASCII(ch)==' ') BF_FORWARD(b);
   }
+  ASSERT_CONSISTENCY(b);
   return 1;
 }
 /*}}}*/
-/* bf_lineend    -- go to end of current line *//*{{{*/
+/* bf_lineend      -- go to end of current line *//*{{{*/
 int bf_lineend(struct Buffer *b, int realend)
 {
-  char c,mark;
+  struct MetaChar ch;
   
-  while (bf_rchar(b,&c,&mark) && c!='\n') BF_FORWARD(b);
+  while (bf_rchar(b,&ch) && ASCII(ch)!='\n') BF_FORWARD(b);
   return 1;
 }
 /*}}}*/
-/* bf_isfold     -- is this a fold? *//*{{{*/
+/* bf_isfold       -- is this a fold? *//*{{{*/
 int bf_isfold(struct Buffer *b, int fold, int where)
 {
-  char *gapstart,c,mark;
+  char *gapstart;
+  struct MetaChar ch;
 
   gapstart=b->gapstart;
   if (where&LEFT)
   {
-    while (bf_lchar(b,&c,&mark))
+    while (bf_lchar(b,&ch))
     {
-      if (c=='\0' && mark&fold)
+      if (ASCII(ch)=='\0' && (MARK(ch)&fold))
       {
         while (b->gapstart<gapstart) BF_FORWARD(b);
         return 1;
       }
-      else if (c=='\n') break;
+      else if (ASCII(ch)=='\n') break;
       if (BF_BACKWARDOK(b)) BF_BACKWARD(b);
     }
     while (b->gapstart<gapstart) BF_FORWARD(b);
   }
   if (where&RIGHT)
   {
-    while (bf_rchar(b,&c,&mark))
+    while (bf_rchar(b,&ch))
     {
-      if (c=='\0' && mark&fold)
+      if (ASCII(ch)=='\0' && (MARK(ch)&fold))
       {
         while (b->gapstart>gapstart) BF_BACKWARD(b);
         return 1;
       }
-      else if (c=='\n') break;
+      else if (ASCII(ch)=='\n') break;
       BF_FORWARD(b);
     }   
     while (b->gapstart>gapstart) BF_BACKWARD(b);
@@ -759,10 +899,11 @@ int bf_isfold(struct Buffer *b, int fold, int where)
   return 0;
 }
 /*}}}*/
-/* bf_foldstart  -- goto start of this fold *//*{{{*/
-int bf_foldstart(struct Buffer *b, int level, unsigned int *linesback)
+/* bf_tofoldstart  -- goto start of this fold *//*{{{*/
+int bf_tofoldstart(struct Buffer *b, int level, int *linesback)
 {
-  char *gapstart,c,mark;
+  char *gapstart;
+  struct MetaChar ch;
   int entrylevel;
   int firstline=level;
         
@@ -771,27 +912,27 @@ int bf_foldstart(struct Buffer *b, int level, unsigned int *linesback)
   if (linesback) *linesback=0;
   do
   {
-    if (!BF_BACKWARDOK(b) || bf_lchar(b,&c,&mark)==0)
+    if (!BF_BACKWARDOK(b) || bf_lchar(b,&ch)==0)
     {
       while (b->gapstart<gapstart) BF_FORWARD(b);
       return 0;
     }
     BF_BACKWARD(b);
-    if (c=='\0' && mark==FOLDEND && !firstline) ++level;
-    else if (c=='\0' && mark==FOLDSTART && !firstline) --level;
-    else if (c=='\n')
+    if (ASCII(ch)=='\0' && MARK(ch)==FOLDEND && !firstline) ++level;
+    else if (ASCII(ch)=='\0' && MARK(ch)==FOLDSTART && !firstline) --level;
+    else if (ASCII(ch)=='\n')
     {
       firstline=0;
       if (level==entrylevel && linesback) ++*linesback;
     }
-  } while (!(c=='\0' && mark==FOLDSTART && level==0));
+  } while (!(ASCII(ch)=='\0' && MARK(ch)==FOLDSTART && level==0));
   return 1;
 }
 /*}}}*/
-/* bf_foldend    -- goto end of this fold *//*{{{*/
-int bf_foldend(struct Buffer *b, int level, unsigned int *linesforward)
+/* bf_tofoldend    -- goto end of this fold *//*{{{*/
+int bf_tofoldend(struct Buffer *b, int level, int *linesforward)
 {
-  char *gapstart,c,mark;
+  char *gapstart;
   int entrylevel;
         
   if (linesforward) *linesforward=0;
@@ -800,33 +941,35 @@ int bf_foldend(struct Buffer *b, int level, unsigned int *linesforward)
   if (!BF_FORWARDOK(b)) return 0;
   do
   {
+    struct MetaChar ch;
+
     BF_FORWARD(b);
-    if (bf_lchar(b,&c,&mark)==0)
+    if (bf_lchar(b,&ch)==0)
     {
       while (b->gapstart>gapstart) BF_BACKWARD(b);
       return 0;
     }
-    if (c=='\0' && mark==FOLDEND) --level;
-    else if (c=='\0' && mark==FOLDSTART) ++level;
-    else if (c=='\n' && level==entrylevel && linesforward) ++*linesforward;
-    if (c=='\0' && mark==FOLDEND && level==0) return 1;
+    if (ASCII(ch)=='\0' && MARK(ch)==FOLDEND) --level;
+    else if (ASCII(ch)=='\0' && MARK(ch)==FOLDSTART) ++level;
+    else if (ASCII(ch)=='\n' && level==entrylevel && linesforward) ++*linesforward;
+    if (ASCII(ch)=='\0' && MARK(ch)==FOLDEND && level==0) return 1;
   } while (BF_FORWARDOK(b));
   return 1;
 }
 /*}}}*/
-/* bf_begin      -- goto begin of buffer *//*{{{*/
+/* bf_begin        -- goto begin of buffer *//*{{{*/
 void bf_begin(struct Buffer *b)
 {
   while (BF_BACKWARDOK(b)) BF_BACKWARD(b);
 }
 /*}}}*/
-/* bf_end        -- goto end of buffer *//*{{{*/
+/* bf_end          -- goto end of buffer *//*{{{*/
 void bf_end(struct Buffer *b)
 {
   while (BF_FORWARDOK(b)) BF_FORWARD(b);
 }
 /*}}}*/
-/* bf_gotoline   -- goto line of buffer *//*{{{*/
+/* bf_gotoline     -- goto line of buffer *//*{{{*/
 int bf_gotoline(struct Buffer *b, int line)
 {
   char *gapstart;
@@ -840,35 +983,34 @@ int bf_gotoline(struct Buffer *b, int line)
   return 0;
 }
 /*}}}*/
-/* bf_strfsearch -- goto line of forwards searched string *//*{{{*/
+/* bf_strfsearch   -- goto line of forward searched string *//*{{{*/
 int bf_strfsearch(struct Buffer *b, const char *needle, size_t needlesize)
 {
   char *found;
 
   if ((b->buffer+b->size-b->gapend-1)>0 && (found=mymemmem(b->gapend+1,b->buffer+b->size-b->gapend-1,needle,needlesize)))
   {
-    while (b->gapend<found) BF_FORWARD(b);
-    if (b->gapend==found) bf_backward(b);
+    while (b->gapend+1<found) BF_FORWARD(b);
     return 1;
   }
   else return 0;
 }
 /*}}}*/
-/* bf_regfsearch -- goto line of forwards searched string *//*{{{*/
+/* bf_regfsearch   -- goto line of forward searched string *//*{{{*/
 int bf_regfsearch(struct Buffer *b, const regex_t *needle)
 {
   regmatch_t found;
   int forward;
   int failed;
-  char c,mark;
+  struct MetaChar ch;
 
   forward=0;
   failed=REG_NOMATCH;
-  while ((failed=regexec(needle,b->gapend+1,1,&found,((b->gapstart==b->buffer || (bf_lchar(b,&c,&mark) && c=='\n')) ? 0 : REG_NOTBOL)|( b->gapend+strlen(b->gapend)<b->buffer+b->size ? REG_NOTEOL : 0 )))==REG_NOMATCH)
+  while ((failed=regexec(needle,b->gapend+1,1,&found,((b->gapstart==b->buffer || (bf_lchar(b,&ch) && ASCII(ch)=='\n')) ? 0 : REG_NOTBOL)|( b->gapend+strlen(b->gapend)<b->buffer+b->size ? REG_NOTEOL : 0 )))==REG_NOMATCH)
   {
-    if (bf_rchar(b,&c,&mark))
+    if (bf_rchar(b,&ch))
     {
-      if (c=='\0')
+      if (ASCII(ch)=='\0')
       {
         bf_forward(b);
         ++forward;
@@ -877,16 +1019,18 @@ int bf_regfsearch(struct Buffer *b, const regex_t *needle)
       {
         bf_forward(b);
         ++forward;
-      } while (bf_rchar(b,&c,&mark) && c!='\0');
+      } while (bf_rchar(b,&ch) && ASCII(ch)!='\0');
     }
     else break;
   }
   if (failed==0)
   {
-    while (found.rm_so>0)
+    if (found.rm_so>0)
     {
-      bf_forward(b);
-      --found.rm_so;
+      const char *rm_so;
+
+      rm_so=b->gapend+1+found.rm_so;
+      while (b->gapend+1<rm_so) BF_FORWARD(b);
     }
     return 1;
   }
@@ -897,7 +1041,7 @@ int bf_regfsearch(struct Buffer *b, const regex_t *needle)
   }
 }
 /*}}}*/
-/* bf_strbsearch -- goto line of backwards searched string *//*{{{*/
+/* bf_strbsearch   -- goto line of backwards searched string *//*{{{*/
 int bf_strbsearch(struct Buffer *b, const char *needle, size_t needlesize)
 {
   int back=0;
@@ -910,17 +1054,17 @@ int bf_strbsearch(struct Buffer *b, const char *needle, size_t needlesize)
   return 1;
 }
 /*}}}*/
-/* bf_regbsearch -- goto line of backwards searched string *//*{{{*/
+/* bf_regbsearch   -- goto line of backwards searched string *//*{{{*/
 int bf_regbsearch(struct Buffer *b, const regex_t *needle)
 {
   regmatch_t found;
   int backward;
   int failed;
-  char c,mark;
+  struct MetaChar ch;
 
   backward=0;
   failed=REG_NOMATCH;
-  while ((failed=regexec(needle,b->gapend+1,1,&found,((b->gapstart==b->buffer || (bf_lchar(b,&c,&mark) && c=='\n')) ? 0 : REG_NOTBOL)|( b->gapend+strlen(b->gapend)<b->buffer+b->size ? REG_NOTEOL : 0 )))==REG_NOMATCH || found.rm_so>0)
+  while ((failed=regexec(needle,b->gapend+1,1,&found,((b->gapstart==b->buffer || (bf_lchar(b,&ch) && ASCII(ch)=='\n')) ? 0 : REG_NOTBOL)|( b->gapend+strlen(b->gapend)<b->buffer+b->size ? REG_NOTEOL : 0 )))==REG_NOMATCH || found.rm_so>0)
   {
     if (bf_backward(b)) ++backward;
     else break;
@@ -936,7 +1080,7 @@ int bf_regbsearch(struct Buffer *b, const regex_t *needle)
   }
 }
 /*}}}*/
-/* bf_tomark     -- goto alternate mark *//*{{{*/
+/* bf_tomark       -- goto alternate mark *//*{{{*/
 void bf_tomark(struct Buffer *b, int to)
 {
   assert(to<b->msz);
@@ -944,12 +1088,12 @@ void bf_tomark(struct Buffer *b, int to)
   while (b->curch>b->mark[to] && BF_BACKWARDOK(b)) BF_BACKWARD(b);
 }
 /*}}}*/
-/* bf_mark       -- set mark *//*{{{*/
+/* bf_mark         -- set mark *//*{{{*/
 int bf_mark(struct Buffer *b, int mark)
 {
   if (mark<0)
   {
-    for (mark=0; b->markused[mark] && mark<b->msz; ++mark);
+    for (mark=0; mark<b->msz && b->markused[mark]; ++mark);
     if (mark==b->msz)
     {
       assert(b->msz<BUFMARKS);
@@ -961,7 +1105,7 @@ int bf_mark(struct Buffer *b, int mark)
   return mark;
 }
 /*}}}*/
-/* bf_unmark     -- unset mark */ /*{{{*/
+/* bf_unmark       -- unset mark *//*{{{*/
 void bf_unmark(struct Buffer *b, int mark)
 {
   assert(mark>=0);
@@ -969,3 +1113,40 @@ void bf_unmark(struct Buffer *b, int mark)
   b->markused[mark]=0;
 }
 /*}}}*/
+/* bf_uc           -- convert meta character to Unicode *//*{{{*/
+long bf_uc(const struct MetaChar *ch)
+{
+  long uc;
+  const char *c, *end;
+
+  assert(ch);
+  c=ch->c;
+  assert(*c);
+  if (!(*c & 0x80))
+  {
+    end=c+1;
+    uc=*c;
+  }
+  else if ((*c & 0xe0) == 0xc0)
+  {
+    end=c+2;
+    uc=*c & 0x1f;
+  }
+  else if ((*c & 0xf0) == 0xe0)
+  {
+    end=c+3;
+    uc=*c & 0xf;
+  }
+  else
+  {
+    assert((*c & 0xf8) == 0xf0);
+    end=c+4;
+    uc=*c & 0x7;
+  }
+  while (++c<end)
+  {
+    uc=(uc<<6)|(*c&0x3f);
+  }
+  return uc;
+}
+/*}}}*/
