@@ -23,29 +23,25 @@
 .ds k \fP
 .\"}}}
 .\"{{{ Title
-.ND "February 10th, 2004"
+.ND "February 10th, 2017"
 .TL
 Fe User Guide
 .AU "Michael Haardt <michael@moria.de>"
 .AF ""
 .AS "" 0.5i
-Yet another editor? Yes, and like all other editors, its special
-design goals may or may not make it your personal favourite.  Fe
-is a folding editor, often also called outline editor.  I believe
-that folding is as important as the progress from line editors to
-screen editors.  Fe is a small editor and needs little resources,
-but still offers many advanced commands.  Instead of an extension
-language, that forces learning yet another programming language
-and bloats the editor, it is designed as a C library containing
-editor primitives and a user interface module.  This module can easily be
-changed if you want extensions or a different user interface.
-Right now there is only one user interface, which is similar to Emacs,
-in particular to the Emacs-like interface of Origami.
+Yet another editor? Yes, and like all other editors, its special design
+goals may or may not make it your personal favourite.  Fe is a folding
+editor, often also called outline editor.  I believe that folding is
+as important as the progress from line editors to screen editors.
+Fe is a small editor and needs little resources, but still offers
+many advanced commands.  It does not contain an extension language,
+but instead is supposed to work out of the box.  The user interface is
+similar to Emacs, in particular to the Emacs-like interface of Origami.
 .AE
 .MT 4 1
 .\"}}}
 .H 1 "Copyright and usage conditions" \"{{{
-\*Cfe\*c (Folding editor) is copyright (c) by Michael Haardt <michael@moria.de>, 1995\(en2004.
+\*Cfe\*c (Folding editor) is copyright (c) by Michael Haardt <michael@moria.de>, 1995\(en2020.
 .P
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -595,9 +591,7 @@ Show the cursor position in the status line.
 .\"{{{ r
 .Li \*Kr\*k
 Read-only.  This mode is automatically set, if you load a file on which
-you have no write permission.  Fe will offer to check in a file to an
-RCS archive, if you set this mode and offer to check it out if you
-delete the mode.
+you have no write permission.
 .\"}}}
 .\"{{{ R
 .Li \*KR\*k
@@ -741,4 +735,4 @@ If you want to transport code from one fold to another, just split the
 current display and edit the file at two places.  If both places are
 sufficient far away from each other, that's what you even had to do if
 you could leave folds open.
-\"}}}
+.\"}}}
