@@ -21,7 +21,7 @@
 /*}}}*/
 
 /* variables *//*{{{*/
-static chtype *mac=(chtype*)0;
+static struct MacroChar *mac=(struct MacroChar*)0;
 static size_t len=0,maxlen=0,runlen=0;
 static int defining=0,runtimes;
 static struct Macro *running=(struct Macro*)0;
@@ -40,14 +40,12 @@ int mc_start(struct Macro *m, int count)
   }
 }
 /*}}}*/
-/* mc_get      -- get character, from macro or from keyboard *//*{{{*/
-chtype mc_get(void)
+/* mc_get      -- get character or function key, from macro or keyboard *//*{{{*/
+void mc_get(struct MacroChar *mch, int utf8term)
 {
-  int ch;
-
   if (running!=(struct Macro*)0 && running->len>runlen)
   {
-    ch=*(running->mac+runlen);
+    *mch=*(running->mac+runlen);
     if (++runlen==running->len)
     {
       if (--runtimes) runlen=0;
@@ -56,17 +54,47 @@ chtype mc_get(void)
   }
   else
   {
+    int key;
+
     (void)refresh();
     /* SIGWINCH may cause one error (interrupted system call, but errno     */
     /* is not set to EINTR after getch(), but never two errors in sequence. */
-    if ((ch=getch())==ERR) ch=getch();
+    if ((key=getch())==ERR) key=getch();
+
+    if (utf8term && (key&~0x7f)==0x80)
+    {
+      /* No function key, but set MSB: Begin of UTF-8 sequence.
+       * Read remaining characters.
+       */
+      int more;
+      char *u;
+
+      u=mch->ch.c;
+      *u=(char)key;
+      more=bf_rcharlen(u)-1;
+      while (more--) *++u=getch();
+      mch->key=KEY_MAX;
+    }
+    else
+    {
+      mch->key=key;
+      if (!(mch->key&~0xff))
+      {
+        ASCII(mch->ch)=(char)key;
+        if (key=='\0')
+        {
+          MARK(mch->ch)=QUOTE;
+          QUOTED(mch->ch)='\0';
+        }
+      }
+    }
+
+    if (defining)
+    {
+      if (len==maxlen) mac=realloc(mac,(maxlen+=256)*sizeof(*mac));
+      *(mac+len++)=*mch;
+    }
   }
-  if (defining)
-  {
-    if (len==maxlen) mac=realloc(mac,(maxlen+=255)*sizeof(chtype));
-    *(mac+len++)=ch;
-  }
-  return ch;
 }
 /*}}}*/
 /* mc_defstart -- start recording a macro *//*{{{*/
@@ -89,7 +117,7 @@ int mc_defend(struct Macro *m, int delete)
     m->len=len-delete;
     if (m->mac) free(m->mac);
     m->mac=mac;
-    mac=(chtype*)0;
+    mac=(struct MacroChar*)0;
     len=0;
     maxlen=0;
     defining=0;
@@ -101,7 +129,7 @@ int mc_defend(struct Macro *m, int delete)
 /* mc_null     -- init macro *//*{{{*/
 void mc_null(struct Macro *m)
 {
-  m->mac=(chtype*)0;
+  m->mac=(struct MacroChar*)0;
   m->len=0;
 }
 /*}}}*/
