@@ -1,4 +1,6 @@
 /* #includes *//*{{{C}}}*//*{{{*/
+#include "config.h"
+
 #ifdef DMALLOC
 #include "dmalloc.h"
 #endif
@@ -33,9 +35,9 @@ of it.
 /*}}}*/
 /*@null@*/ char *mymemmem(const char *hay, size_t haysize, const char *needle, size_t needlesize)
 {
-  while (haysize!=0)
+  while (haysize)
   {
-    while (haysize!=0 && *needle!=*hay) { ++hay; --haysize; }
+    while (haysize && *needle!=*hay) { ++hay; --haysize; }
     if (haysize==0) return (char*)0;
     else
     {
@@ -98,8 +100,9 @@ of it.
   return (char*)0;
 }
 /*}}}*/
-/* mystrmalloc -- return malloced copy of string */ /*{{{*/
-/*@null@*/ char *mystrmalloc(const char *str)
+#ifndef HAVE_STRDUP
+/* strdup      -- return malloced copy of string */ /*{{{*/
+/*@null@*/ char *strdup(const char *str)
 {
   char *m;
 
@@ -108,8 +111,26 @@ of it.
   return m;
 }
 /*}}}*/
+#endif
+#ifndef HAVE_STRNDUP
+/* strndup     -- return malloced copy of string */ /*{{{*/
+/*@null@*/ char *strndup(const char *str, size_t len)
+{
+  char *m;
+  const char *r,*end;
+
+  for (r=str,end=r+len; *r && r<end; ++r);
+  len=r-str;
+
+  if ((m=malloc(len+1))==(char*)0) return (char*)0;
+  memcpy(m,str,len);
+  *(m+len)='\0';
+  return m;
+}
+/*}}}*/
+#endif
 /* myrealloc   -- ANSI conforming realloc() */ /*{{{*/
-#ifdef OLD_REALLOC
+#ifdef BROKEN_REALLOC
 #undef realloc
 void *myrealloc(void *p, size_t n)
 {
@@ -134,7 +155,7 @@ void *myrealloc(void *p, size_t n)
       }
       else
       {
-        shell=mystrmalloc(pwd->pw_shell);
+        shell=strdup(pwd->pw_shell);
       }
     }
   }
@@ -150,10 +171,22 @@ void *myrealloc(void *p, size_t n)
 
   if ((tmpdir=getenv("TMPDIR"))==(char*)0) tmpdir="/tmp";
   if ((strlen(tmpdir)+1+8+1)>(size_t)L_tmpnam) return (char*)0;
-  i=getpid();
-  while (i<0xffffffff && snprintf(buf,(size_t)L_tmpnam,"%s/%08x",tmpdir,(unsigned int)i)!=0,(fd=open(buf,O_RDWR|O_CREAT|O_EXCL,0666))==-1 && errno==EEXIST) ++i;
-  if (fd==-1) return (char*)0;
-  (void)close(fd);
-  return buf;
+  buf[0]='\0';
+  for (i=getpid(); i; ++i)
+  {
+    snprintf(buf,(size_t)L_tmpnam,"%s/%08x",tmpdir,(unsigned int)i);
+    if (((fd=open(buf,O_RDWR|O_CREAT|O_EXCL,0666))!=-1) || errno!=EEXIST) break;
+    if (fd!=-1) (void)close(fd);
+  }
+  if (fd!=-1)
+  {
+    (void)close(fd);
+    return buf;
+  }
+  else
+  {
+    fprintf(stderr,"fail '%s' (%s)\n",buf,strerror(errno));
+    return (char*)0;
+  }
 }
 /*}}}*/
