@@ -4,7 +4,11 @@
 #include <sys/types.h>
 #include <stdarg.h> /* required by some curses.h implementations */
 #ifdef NEED_NCURSES
+#ifdef HAVE_NCURSES_NCURSES_H
+#include <ncurses/ncurses.h>
+#else
 #include <ncurses.h>
+#endif
 #else
 #include <curses.h>
 #endif
@@ -14,11 +18,11 @@
 typedef struct
 {
   chtype c;
-  char *str;
+  const char *str;
 } MenuChoice;
 
 void showmsg(const char *fmt, ...);
-int editline(char *buf, size_t size, const char *prompt, size_t *x, size_t *offx, char *history[HISTORY_LEN]);
+char *editline(const char *prompt, const char *preset, char *history[HISTORY_LEN]);
 int menuline(const char *prompt, const MenuChoice *choice, int curx);
 int ok(const char *prompt, int curx);
 
